package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.InvalidEventLog;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.steps.subscriber.RequestSubmissionStep;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

class SubscriberClientTest {

    private static final ObjectMapper mapper = new ObjectMapper();
    private final LogCaptor logs = new LogCaptor();

    private Subscriber subscriberMock;
    private RSAKey privateKey;
    private RequestSubmissionStep underTest;

    @BeforeEach
    public void setup() throws IOException, ParseException {
        this.privateKey = RSAKey.parse(getResourceAsString("private_decryption_test_key.json"));
        this.subscriberMock = Mockito.mock(Subscriber.class);
        this.underTest = SubscriberClient.builder(this.subscriberMock, this.privateKey);
    }

    @Test
    void testMissingPrivateKey() {
        assertNull(SubscriberClient.builder(this.subscriberMock, null));
        this.logs.assertContains("No decryption key set, please check environment");
    }

    @Test
    void testNoSubmissionsFound() {
        assertThat(this.underTest.getAvailableSubmissions(null).isEmpty(), is(true));
        assertThat(this.underTest.getAvailableSubmissions(UUID.randomUUID()).isEmpty(), is(true));

        final var destinationId = UUID.randomUUID();
        when(this.subscriberMock.pollAvailableSubmissions(destinationId, 0, 5)).thenReturn(Collections.emptySet());
        assertThat(this.underTest.getAvailableSubmissions(destinationId).isEmpty(), is(true));
    }

    @Test
    void testRestCallFailed() {
        // Given
        when(this.subscriberMock.getSubmission(any())).thenThrow(new RestApiException("Submission not found"));

        // When
        final var receivedSubmission = this.underTest.requestSubmission(UUID.randomUUID()).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isEmpty());
        this.logs.assertContains("API request failed");
    }

    @Test
    void testDecryptionFailed() {
        // Given
        final var submissionId = UUID.randomUUID();
        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData("encryptedData");
        submission.setEncryptedMetadata("encryptedMetadata");

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.decryptStringContent(any(), any())).thenThrow(new DecryptionException("Decryption failed"));

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isEmpty());
        this.logs.assertContains("Decrypting payload failed");
    }

    @Test
    void testAcceptSubmission() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setCaseId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());

        final var hash = new Hash();
        hash.setContent("hashedTestContent");

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
        submissionSchema.setSchemaUri(URI.create("https://test-schema.fitko.de"));

        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.decryptStringContent(any(), any())).thenReturn(mapper.writeValueAsBytes(metadata));
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.ok());

        // When
        this.underTest.requestSubmission(submissionId).acceptSubmission();

        // Then
        verify(this.subscriberMock, times(1)).acceptSubmission(new EventPayload(submission));
    }

    @Test
    void testUserRejectSubmission() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setCaseId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.decryptStringContent(any(), any())).thenReturn(mapper.writeValueAsBytes("metadata"));
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.error(new ValidationException("invalid metadata")));

        // When
        this.underTest.requestSubmission(submissionId).rejectSubmission(List.of(new InvalidEventLog()));

        // Then
        final var expectedEventPayload = new EventPayload(submission, List.of(new InvalidEventLog()));
        verify(this.subscriberMock, times(1)).rejectSubmission(expectedEventPayload);
    }

    @Test
    void testReadingMetadataFailed() {
        // Given
        final var submissionId = UUID.randomUUID();
        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData("encryptedData");
        submission.setEncryptedMetadata("encryptedMetadata");

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.decryptStringContent(any(), any())).thenReturn("encryptedMetadata".getBytes());

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isEmpty());
        this.logs.assertContains("Reading metadata failed");
    }

    @Test
    void testSubmissionsFound() {
        // Given
        final var destinationId = UUID.randomUUID();

        final SubmissionForPickup submission = new SubmissionForPickup();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(destinationId);

        when(this.subscriberMock.pollAvailableSubmissions(destinationId, 0, 1)).thenReturn(Set.of(submission));

        // When
        final Set<SubmissionForPickup> availableSubmissions = this.underTest.getAvailableSubmissions(destinationId, 0, 1);

        // Then
        assertThat(availableSubmissions.size(), is(1));
        assertThat(availableSubmissions.stream().findFirst().get(), is(submission));
    }

    @Test
    void testGetSubmission() throws Exception {

        // Given
        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Hash hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        hash.setSignatureType(SignatureType.SHA_512);

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://dummy.schema.url"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of());
        contentStructure.setData(data);

        final Metadata metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final byte[] metadataBytes = mapper.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = this.privateKey;
        final RSAKey decryptionKey = encryptionKey.toPublicJWK();

        final String encryptedData = cryptoService.encryptString(decryptionKey, dataPayload);
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes);

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(this.subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isPresent());
        assertThat(receivedSubmission.get().getData(), is(dataPayload));
        assertThat(receivedSubmission.get().getDataMimeType(), is(MimeType.APPLICATION_JSON));
        assertThat(receivedSubmission.get().getAttachments(), is(Collections.EMPTY_LIST));
    }

    @Test
    void testSubmissionNotFound() {
        // Given
        when(this.subscriberMock.getSubmission(any())).thenThrow(new RestApiException("Submission not found"));

        // When
        final var receivedSubmission = this.underTest.requestSubmission(UUID.randomUUID()).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isEmpty());
    }

    @Test
    void testInvalidMetadata() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setCaseId(caseId);

        final var invalidMetadata = mapper.writeValueAsBytes(new Metadata());

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.decryptStringContent(any(), any())).thenReturn(invalidMetadata);
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.error(new ValidationException("Validation failed")));

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isEmpty());
        this.logs.assertContains("Metadata does not match schema");
    }

    @Test
    void testCorruptedData() throws JsonProcessingException {

        // Given
        final RSAKey decryptionKey = this.privateKey;
        final RSAKey encryptionKey = decryptionKey.toPublicJWK();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final var dataPayload = "some data that was transferred";

        final var hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        final var data = new Data();
        data.setHash(hash);
        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String encryptedData = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(data));
        final String encryptedMetadata = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(metadata));
        final var metadataBytes = mapper.writeValueAsBytes(metadata);
        final var dataBytes = mapper.writeValueAsBytes(data);

        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setCaseId(caseId);
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setEncryptedData(encryptedData);

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.decryptStringContent(decryptionKey, encryptedMetadata)).thenReturn(metadataBytes);
        when(this.subscriberMock.decryptStringContent(decryptionKey, encryptedData)).thenReturn(dataBytes);
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.error(new ValidationException("Corrupt data hash")));

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isEmpty());
        this.logs.assertContains("Data might be corrupted, hash-sum does not match");
    }

    @Test
    void testCorruptedAttachment() throws JsonProcessingException {
        // Given
        final RSAKey decryptionKey = this.privateKey;
        final RSAKey encryptionKey = decryptionKey.toPublicJWK();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final var dataHash = new Hash();
        dataHash.setContent("");
        final var data = new Data();
        data.setHash(dataHash);

        final var attachmentHash = new Hash();
        attachmentHash.setContent("");
        final var attachment = new Attachment();
        attachment.setFilename("src/test/resources/attachment.txt");
        attachment.setHash(attachmentHash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(List.of(attachment));
        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String encryptedData = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(data));
        final String encryptedMetadata = cryptoService.encryptString(encryptionKey, mapper.writeValueAsString(metadata));
        final var metadataBytes = mapper.writeValueAsBytes(metadata);
        final var dataBytes = mapper.writeValueAsBytes(data);

        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setCaseId(caseId);
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setEncryptedData(encryptedData);

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.decryptStringContent(decryptionKey, encryptedMetadata)).thenReturn(metadataBytes);
        when(this.subscriberMock.decryptStringContent(decryptionKey, encryptedData)).thenReturn(dataBytes);
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.ok());

        final var decryptedDataBytes = cryptoService.decryptBytes(decryptionKey, submission.getEncryptedData());
        when(this.subscriberMock.validateHashIntegrity(dataHash.getContent(), decryptedDataBytes)).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.validateHashIntegrity(attachmentHash.getContent(), null)).thenReturn(ValidationResult.error(new ValidationException("Corrupt attachment")));

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isEmpty());
        this.logs.assertContains("Attachment data for id null is corrupted");
    }

    @Test
    void testGetValidEventLog() {

        // Given
        final var caseId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final List<EventLogEntry> expectedLog = List.of(EventLogEntry.builder().build());
        when(this.subscriberMock.getEventLog(any(), any())).thenReturn(expectedLog);

        // When
        final List<EventLogEntry> eventLog = this.underTest.getEventLog(caseId, destinationId);

        // Then
        assertThat(eventLog, is(expectedLog));

    }

    @Test
    void testGetInvalidEventLog() {

        // Given
        final var caseId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        when(this.subscriberMock.getEventLog(any(), any())).thenThrow(new EventLogException("Invalid log"));

        // When
        final EventLogException exception = assertThrows(EventLogException.class, () -> this.underTest.getEventLog(caseId, destinationId));

        // Then
        assertThat(exception.getMessage(), containsString("Invalid log"));

    }

    @Test
    void testUnknownAdditionalReferenceInfoProperty() throws JsonProcessingException {

        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Map<String, Map<String, Object>> metadata = Map.of(
                "contentStructure", Map.of(
                        "data", Map.of(
                                "hash", Map.of(
                                        "type", SignatureType.SHA_512,
                                        "content", cryptoService.hashBytes(dataPayload.getBytes())
                                ),
                                "submissionSchema", Map.of(
                                        "schemaUri", URI.create("https://dummy.schema.url"),
                                        "mimeType", MimeType.APPLICATION_JSON
                                )
                        ),
                        "attachments", Collections.EMPTY_LIST),
                "additionalReferenceInfo", Map.of(
                        "x-sender", "SOME_ADDITIONAL_UNKNOWN_PROPERTY",
                        "senderReference", "sender_123")
        );

        final byte[] metadataBytes = mapper.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = this.privateKey;
        final RSAKey decryptionKey = encryptionKey.toPublicJWK();

        final String encryptedData = cryptoService.encryptString(decryptionKey, dataPayload);
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes);

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(this.subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isPresent());
    }

    @Test
    void testUnknownSchemaPropertyOnMetadata() throws JsonProcessingException {

        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Map<String, Map<String, Object>> metadata = Map.of(
                "$schema", Collections.emptyMap(),
                "contentStructure", Map.of(
                        "data", Map.of(
                                "hash", Map.of(
                                        "type", SignatureType.SHA_512,
                                        "content", cryptoService.hashBytes(dataPayload.getBytes())
                                ),
                                "submissionSchema", Map.of(
                                        "schemaUri", URI.create("https://dummy.schema.url"),
                                        "mimeType", MimeType.APPLICATION_JSON
                                )
                        )
                )
        );

        final byte[] metadataBytes = mapper.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = this.privateKey;
        final RSAKey decryptionKey = encryptionKey.toPublicJWK();

        final String encryptedData = cryptoService.encryptString(decryptionKey, dataPayload);
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes);

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        when(this.subscriberMock.getSubmission(any())).thenReturn(submission);
        when(this.subscriberMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.validateHashIntegrity(any(), any())).thenReturn(ValidationResult.ok());
        when(this.subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(this.subscriberMock.decryptStringContent(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);

        // When
        final var receivedSubmission = this.underTest.requestSubmission(submissionId).getSubmissionData();

        // Then
        assertTrue(receivedSubmission.isPresent());
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        final File file = new File(Objects.requireNonNull(classLoader.getResource(name)).getFile());
        return Files.readString(file.toPath());
    }
}